﻿//////////////////////////////////////////////
// FreeListMemorySpaceWrapper.h
//
//////////////////////////////////////////////

/// Defines / Macros -------------------------

#pragma once

/// Includes ---------------------------------

// nkMemory
#include "../../Log/LogManager.h"

#include "../MemorySpaces/MemorySpace.h"

// Standards
#include <map>

/// Internals --------------------------------

namespace nkMemory
{
	// Size_t codant pour invalide
	constexpr size_t FREE_LIST_INVALID_SIZE = (size_t)-1 ;

	// Info de spots libres (map de spaces)
	struct FreeListFreeSpot
	{
		size_t _size ;
	} ;
}

/// Class ------------------------------------

namespace nkMemory
{
	template <typename T, typename U>
	class FreeListMemorySpaceWrapper final
	{
		public :

			// Attributs
			U _memSpace ;

			// La map des free spots pour ce space, par offset
			std::map<size_t, FreeListFreeSpot> _freeSpots ;

			// Taille restante totale dans le space
			size_t _freeSize ;

		public :

			// Track spot
			inline void _addFreeSpot (size_t offset, size_t size)
			{
				_freeSpots.emplace(offset, FreeListFreeSpot{size}) ;
				_freeSize += size ;
			}

			// Recherche de bloc
			inline std::map<size_t, FreeListFreeSpot>::const_iterator _searchForPrecedingFreeSpot (size_t offset) const
			{
				std::map<size_t, FreeListFreeSpot>::const_iterator searchResult = _freeSpots.lower_bound(offset) ;

				// Check bounds
				if (searchResult == _freeSpots.end() || searchResult->first > offset)
					searchResult-- ;

				return searchResult ;
			}

			inline std::map<size_t, FreeListFreeSpot>::const_iterator _searchForFollowingFreeSpot (size_t offset) const
			{
				std::map<size_t, FreeListFreeSpot>::const_iterator searchResult = _freeSpots.upper_bound(offset) ;

				return searchResult ;
			}

			// Split bloc

		public :

			// Constructeur
			FreeListMemorySpaceWrapper (size_t spaceSize)
			:	_memSpace (spaceSize),
				_freeSpots (),
				_freeSize (0)
			{
				// Ajout d'un free spot sur la totalité du space
				_addFreeSpot(0, spaceSize) ;
			}

			// Destructeur
			~FreeListMemorySpaceWrapper () = default ;

		public :

			// Getters
			size_t getFreeSize () const
			{
				return _freeSize ;
			}

		public :

			// Demande d'allocation
			size_t getFittingOffsetFor (size_t size, size_t alignment) const
			{
				// Quick check
				if (size > _freeSize)
					return FREE_LIST_INVALID_SIZE ;

				// On retrouve un free spot compatible
				size_t allocationOffset = 0 ;
				size_t toAlignOffset = 0 ;
				size_t alignedOffset = 0 ;

				for (const std::pair<const size_t, const FreeListFreeSpot>& entry : _freeSpots)
				{
					allocationOffset = entry.first ;
					toAlignOffset = allocationOffset % alignment ;
					alignedOffset = toAlignOffset ? allocationOffset + (alignment - toAlignOffset) : allocationOffset ;

					if (alignedOffset + size - allocationOffset <= entry.second._size)
						return alignedOffset ;
				}

				return FREE_LIST_INVALID_SIZE ;
			}

			void noticeAllocation (size_t offset, size_t size)
			{
				// On retrouve le free spot compatible
				std::map<size_t, FreeListFreeSpot>::const_iterator searchResult = _searchForPrecedingFreeSpot(offset) ;

				// En tout logique il ne sera pas possible de ne pas trouver à cause des interactions avec le Pager

				// Split du block donc
				// Temporaires
				size_t spotOffset = searchResult->first ;
				FreeListFreeSpot spot = searchResult->second ;

				// Clear de l'entrée multi map
				_freeSpots.erase(searchResult) ;

				// Si alignement offset, alors un nouveau bloc à gauche
				size_t spotInternalOffset = offset - spotOffset ;

				if (spotInternalOffset)
					_addFreeSpot(spotOffset, spotInternalOffset) ;

				// Si reste à droite maintenant
				size_t totalSize = spotInternalOffset + size ;

				if (totalSize < spot._size)
					_addFreeSpot(totalSize + spotOffset, spot._size - totalSize) ;

				// Update free size
				_freeSize -= spot._size ;
			}

			void noticeFree (size_t offset, size_t size)
			{
				// Vide ?
				if (_freeSpots.empty())
				{
					_addFreeSpot(offset, size) ;
				}
				else
				{
					// Infos finale de nouveau bloc
					size_t finalOffset = offset ;
					size_t finalSize = size ;

					// Spot à gauche ?
					std::map<size_t, FreeListFreeSpot>::const_iterator searchResult = _searchForPrecedingFreeSpot(offset) ;

					if (searchResult != _freeSpots.end())
					{
						size_t spotEndingOffset = searchResult->first + searchResult->second._size ;

						// Merge ?
						if (spotEndingOffset == offset)
						{
							finalOffset = searchResult->first ;
							finalSize += searchResult->second._size ;
							_freeSize -= searchResult->second._size ;

							_freeSpots.erase(searchResult) ;
						}
					}

					// Spot à droite ?
					searchResult = _searchForFollowingFreeSpot(offset) ;

					if (searchResult != _freeSpots.end())
					{
						size_t spotEndingOffset = offset + size ;

						// Merge ?
						if (spotEndingOffset == searchResult->first)
						{
							finalSize += searchResult->second._size ;
							_freeSize -= searchResult->second._size ;

							_freeSpots.erase(searchResult) ;
						}
					}

					// Et on recrée le bloc
					_addFreeSpot(finalOffset, finalSize) ;
				}
			}

		public :

			// Utilitaires
			T getOffsetPtr (size_t offset) const
			{
				return _memSpace.getOffsetPtr(offset) ;
			}
	} ;
}